@@ -8,7 +8,7 @@
  work, as well as functions which operate on them.
 """
 from types import FunctionType as Function
-from typing import Any, Dict, Generic, Hashable, TypeVar
+from typing import Any, Dict, Generic, Hashable, List, Type, TypeVar, Union
 
 DataClass = Any  # type hint for variables that should be data class instances
 T = TypeVar('T')
@@ -18,9 +18,9 @@ class Internal(Generic[T]):
     """This type hint wrapper represents a field that is internal to the data class and is, for example, not to be
     shown in a repr."""
     @classmethod
-    def is_internal(cls, annotation):
+    def is_internal(cls, annotation: Union[Type, str]):
         return getattr(annotation, '__origin__', None) is cls or \
-               (type(annotation) is str and cls.__name__ in annotation)
+               (type(annotation) is str and annotation.startswith(f'{cls.__name__}['))
 
 
 class DataClassMeta(type):
@@ -43,6 +43,10 @@ class DataClassMeta(type):
             options.update(b.get('__dataclass__', {}))
         options.update(kwargs)
 
+        # determine whether a post-initialisation method is defined
+        post_init = (type(dict_.get('__init__')) is Function and not dataclass_bases
+                     or type(dict_.get('__post_init__')) is Function)
+
         # fill out this class' dict and store defaults, annotations and decorator options for future subclasses
         dict_.update(all_defaults)
         dict_['__defaults__'] = all_defaults
@@ -53,108 +57,96 @@ class DataClassMeta(type):
         del dict_['__dict__'], dict_['__weakref__']
 
         # create/apply generated methods and attributes
-        user_init = type(dict_.get('__init__')) is Function
-
         if options['slots']:
-            # values with default values must only be present in slots, not dict, otherwise Python will interpret them
-            # as read only
+            # if the slots option is added, specify them. Values with default values must only be present in slots,
+            # not dict, otherwise Python will interpret them as read only
             for d in all_annotations.keys() & all_defaults.keys():
                 del dict_[d]
             dict_['__slots__'] = all_annotations.keys() - all_slots
         elif '__slots__' in dict_:
-            # if the slots option has been removed from an inheriting dataclass we must remove descriptors and __slots__
+            # if the slots option gets removed, remove descriptors and __slots__
             for d in all_annotations.keys() - all_defaults.keys() & dict_.keys():
                 del dict_[d]
             del dict_['__slots__']
-        if options['init']:
-            dict_.setdefault('__new__', _generate_new(all_annotations, all_defaults, user_init,
-                                                      options['kwargs'], options['frozen']))
+
+        if options['init'] and all_annotations:  # only generate __init__ if there are fields to set
+            if post_init and '__init__' in dict_:
+                dict_['__post_init__'] = dict_.pop('__init__')
+            dict_['__init__'] = generate_init(all_annotations, all_defaults, post_init,
+                                              options['kwargs'], options['frozen'])
+
         if options['repr']:
             dict_.setdefault('__repr__', __repr__)
+
         if options['eq']:
             dict_.setdefault('__eq__', __eq__)
+
         if options['iter']:
             dict_.setdefault('__iter__', __iter__)
+
         if options['frozen']:
             dict_['__delattr__'] = dict_['__setattr__'] = __setattr__
+
         if options['order']:
             dict_.setdefault('__lt__', __lt__)
+
         if (options['eq'] and options['frozen']) or options['unsafe_hash']:
             dict_.setdefault('__hash__', __hash__)
 
-        return type.__new__(mcs if not user_init else DataClassInit, name, bases, dict_)
+        return type.__new__(mcs, name, bases, dict_)
 
+    # noinspection PyMissingConstructor,PyUnresolvedReferences,PyTypeChecker,PyUnusedLocal
     def __init__(cls, *args, **kwargs):
-        # warn the user if they try to use __post_init__
-        if hasattr(cls, '__post_init__'):
-            raise TypeError('dataclassy does not use __post_init__. You should rename this method __init__')
-
         if cls.__dataclass__['eq'] and cls.__dataclass__['order']:
             from functools import total_ordering
             total_ordering(cls)
 
         # determine a static expression for an instance's fields as a tuple, then evaluate this to create a property
         # allowing efficient representation for internal methods
-        tuple_expr = f'({", ".join((*(f"self.{f}" for f in fields(cls)), ""))})'  # "" ensures closing comma
-        cls.__tuple__ = property(eval(f'lambda self: {tuple_expr}'))
+        tuple_expr = ', '.join((*(f'self.{f}' for f in fields(cls)), ''))  # '' ensures closing comma
+        cls.__tuple__ = property(eval(f'lambda self: ({tuple_expr})'))
 
 
-class DataClassInit(DataClassMeta):
-    """In the case that a custom __init__ is defined, remove arguments used by __new__ before calling it."""
-    def __call__(cls, *args, **kwargs):
-        args = iter(args)
-        new_kwargs = dict(zip(cls.__annotations__, args))  # convert positional args to keyword for __new__
-        instance = cls.__new__(cls, **new_kwargs, **kwargs)
-
-        for parameter in kwargs.keys() & cls.__annotations__.keys():
-            del kwargs[parameter]
-
-        instance.__init__(*args, **kwargs)
-        return instance
-
-    @property
-    def __signature__(cls):
-        """Defining a __call__ breaks inspect.signature. Lazily generate a Signature object ourselves."""
-        import inspect
-        parameters = tuple(inspect.signature(cls.__new__).parameters.values())
-        return inspect.Signature(parameters[1:])  # remove 'cls' to transform parameters of __new__ into those of class
-
-
-def _generate_new(annotations: Dict, defaults: Dict, user_init: bool, gen_kwargs: bool, frozen: bool) -> Function:
-    """Generate and return a __new__ method for a data class which has as parameters all fields of the data class.
-    When the data class is initialised, arguments to this function are applied to the fields of the new instance. Using
-    __new__ frees up __init__, allowing it to be defined by the user to perform additional, custom initialisation.
-
-    If __init__ is defined, all arguments to __new__ become keyword-only, and the custom __call__ converts positional
-    arguments to keyword arguments. This prevents ambiguity during attempts to pass the same argument twice,
-    positionally and as a keyword."""
+def generate_init(annotations: Dict, defaults: Dict, user_init: bool, gen_kwargs: bool, frozen: bool) -> Function:
+    """Generate and return an __init__ method for a data class. This method has as parameters all fields of the data
+    class. When the data class is initialised, arguments to this function are applied to the fields of the new instance.
+    A user-defined __init__, if present, must be aliased to avoid conflicting."""
     arguments = [a for a in annotations if a not in defaults]
     default_arguments = [f'{a}={a}' for a in annotations if a in defaults]
-    kw_only = ['*'] if user_init and (arguments or default_arguments) else []
-    kwargs = ['**kwargs'] if gen_kwargs or user_init else []  # pass through unknown keyword arguments to __init__
+    args = ['*args'] if user_init else []
+    kwargs = ['**kwargs'] if user_init or gen_kwargs else []
 
-    parameters = ', '.join(kw_only + arguments + default_arguments + kwargs)
+    parameters = ', '.join(arguments + default_arguments + args + kwargs)
+
+    # surprisingly, given global lookups are slow, using them is the fastest way to compare a field to its default
+    # the alternatives are to look up on self (which wouldn't work when slots=True) or look up self.__defaults__
+    default_names = {f'default_{n}': v for n, v in defaults.items()}
 
     # determine what to do with arguments before assignment. If the argument matches a mutable default, make a copy
-    references = {n: f'{n}.copy() if {n} is self.__defaults__[{n!r}] else {n}'
-                  if n in defaults and hasattr(defaults[n], 'copy') else n for n in annotations}
+    references = {n: f'{n}.copy() if {n} is default_{n} else {n}' if n in defaults and hasattr(defaults[n], 'copy')
+                  else n for n in annotations}
 
-    # if the class is frozen, use the necessary but slightly slower object.__setattr__
-    assignments = [f'object.__setattr__(self, {n!r}, {r})' if frozen else f'self.{n} = {r}'
-                   for n, r in references.items()]
+    # if the class is frozen, use the necessary but far slower object.__setattr__
+    assignments = [f'object.__setattr__(self, {n!r}, {r})' if frozen
+                   else f'self.{n} = {r}' for n, r in references.items()]
 
     # generate the function
-    signature = f'def __new__(cls, {parameters}):'
-    body = [f'self = object.__new__(cls)', *assignments, 'return self']
+    lines = [f'def __init__(self, {parameters}):',
+             *assignments,
+             'self.__post_init__(*args, **kwargs)' if user_init else '']
 
-    exec('\n\t'.join([signature, *body]), {}, defaults)
-    function = defaults.pop('__new__')
+    return eval_function('__init__', lines, annotations, defaults, default_names)
+
+
+def eval_function(name: str, lines: List[str], annotations: Dict, locals_: Dict, globals_: Dict) -> Function:
+    """Evaluate a function definition, returning the resulting object."""
+    exec('\n\t'.join(lines), globals_, locals_)
+    function = locals_.pop(name)
     function.__annotations__ = annotations
     return function
 
 
 # generic method implementations common to all data classes
-# these are currently relatively inefficient - it would be better to cache an expression for a class' tuple
 from .functions import values, fields
 
 
@@ -170,7 +162,7 @@ def __lt__(self: DataClass, other: DataClass):
 
 def __hash__(self):
     # currently not ideal since gives no warning if a field expected to be hashable is not
-    return hash((type(self), *(v for v in self.__tuple__ if v is isinstance(v, Hashable))))
+    return hash((type(self), *(v for v in self.__tuple__ if isinstance(v, Hashable))))
 
 
 def __iter__(self):
@@ -179,9 +171,11 @@ def __iter__(self):
 
 def __repr__(self):
     show_internals = not self.__dataclass__['hide_internals']
-    field_values = ', '.join(f'{f}={v!r}' for f, v in values(self, show_internals).items())
+    field_values = ', '.join(f'{f}=...' if v is self
+                             else f'{f}={v!r}' for f, v in values(self, show_internals).items())
     return f'{type(self).__name__}({field_values})'
 
 
+# noinspection PyUnusedLocal
 def __setattr__(self, *args):
     raise AttributeError('Frozen class')
